html,body {
    height:100%;
    background-color: #fafafa;
}

i[class*="icon"] {
  margin-right:5px;
  line-height: 1;
}

.terminal {
	display: block;
	background: #000; 
	color: #fff; 
	overflow: auto; 
	width: 65%; 
	height: 250px;
	max-height: 250px;
	width: 100%;
	text-align: left;
}

.invisible {
    visibility: hidden;
}

.navbar-grey {
	color: #fff;
	background-color: #464646;
}

.node-info-body {
	border-top: 5px solid #AD3636; 
	border-bottom: 4px solid #a6a6a6; 
	background-color: #d7d7eb; 
	/*position: relative; top: -22px*/
}

.node-body-heading {
	padding-bottom: 5px; 
	border-bottom: 3px solid #a6a6a6	
}

.node-headline-big {
 font-size: 19px;	
}

.gp-container-header {
    /*
    background-color: transparent;
    background-image: linear-gradient(to bottom, rgba(30, 87, 153, 0.1) 0%, rgba(125, 185, 232, 0) 100%);
    background-repeat: repeat;
    */
    padding: 5px;
    text-align: center;
    border-bottom: 1px solid #e0e0e0;
}

.gp-container {
    border: 1px solid lightgrey;
    border-radius:5px;
    text-align: center;
}

.gp-container-content {
    padding: 5px;
    text-align: center;
}