<!DOCTYPE html>
<html lang="en">
<head>
    <title>{{ escape(page_title) }}</title>
    <meta charset="utf-8">
    <meta http-equiv="X-UA-Compatible" content="IE=edge">
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <!-- stylesheets -->
    <link rel="stylesheet" href="css/bootstrap.min.css">
    <link rel="stylesheet" href="https://netdna.bootstrapcdn.com/font-awesome/4.3.0/css/font-awesome.css">
    <link rel="stylesheet" href="css/gambolputty_web.css">
</head>
<body>
<div class="container-fluid">
    <div class="row">
        <div class="col-md-12">
            <nav class="navbar navbar-default navbar-inverse" role="navigation">
                <div class="navbar-header">

                    <button type="button" class="navbar-toggle" data-toggle="collapse"
                            data-target="#bs-example-navbar-collapse-1">
                        <span class="sr-only">Toggle navigation</span><span class="icon-bar"></span><span
                            class="icon-bar"></span><span class="icon-bar"></span>
                    </button>
                    <a class="navbar-brand" href="#">Gambolputty</a>
                </div>

                <div class="collapse navbar-collapse" id="bs-example-navbar-collapse-1">
                    <ul class="nav navbar-nav">
                        <li class="active">
                            <a href="#"><i class="fa fa-sitemap"></i> pack</a>
                        </li>
                        <li>
                            <a href="#"><i class="fa fa-server"></i> members</a>
                        </li>
                    </ul>
                    <form class="navbar-form navbar-right" role="search">
                        <div class="form-group">
                            <input type="text" class="form-control">
                        </div>
                        <button type="submit" class="btn btn-default">
                            Submit
                        </button>
                    </form>
                </div>
            </nav>
        </div>
    </div>
    <div class="row">
        <div class="col-md-2">
            <div class="gp-container">
                <div class="gp-container-header">
                    LogPack
                </div>
                <div class="gp-container-content">
                    <i class="fa fa-server"></i>&nbsp;<b>{{ len(nodes) }}</b>&nbsp;member(s)&nbsp;&nbsp;<i class="fa fa-bar-chart"></i>&nbsp;<b>2015</b>&nbsp;events/s
                </div>
            </div>
        </div>
        <div class="col-md-2">
            {% for node in nodes %}
                {% module ServerInfo(node) %}
            {% end %}
        </div>
        <div class="col-md-8">
            <div class="gp-container">
                <div class="gp-container-header">
                    <div class="row">
                        <div class="col-md-2" style="text-align: left">
                            <div class="btn-group" style="text-align: left">
                                <a class="btn btn-default" data-toggle="dropdown" href="#">displaying logs</a><a class="btn btn-primary dropdown-toggle" data-toggle="dropdown" href="#"><span class="caret"></span></a>
                                <ul class="dropdown-menu">
                                    <li>
                                        <a href="#" onClick=""><i class="fa fa-list pull-left icon"></i><span>&nbsp;show&nbsp;logs</span></a>
                                    </li>
                                    <li>
                                        <a href="#" onClick="showServerConfiguration()"><i class="fa fa-sliders pull-left icon"></i><span>&nbsp;show&nbsp;configuration&nbsp;(text)</span></a>
                                    </li>
                                    <li>
                                        <a href="#"><i class="fa fa-sliders pull-left icon"></i><span>&nbsp;show&nbsp;configuration&nbsp;(graph)</span></a>
                                    </li>

                                </ul>
                            </div>
                        </div>
                        <div class="col-md-10">
                        </div>
                    </div>
                </div>
                <div class="gp-container-content">
                    <div class="terminal" id="{{ escape(node['server_name']) }}_log"></div>
                </div>
            </div>
        </div>
    </div>
</div>
<!-- scripts -->
<script type="text/javascript" src="js/jquery.min.js"></script>
<script type="text/javascript" src="js/bootstrap.min.js"></script>
<script type="text/javascript" src="js/config.js"></script>
<script type="text/javascript" src="js/utils.js"></script>
<script type="text/javascript" src="js/gambolputty_web.js"></script>
</body>
</html>