/**
 * Some configuration and global settings
 */
var globalSettings = {
	"serverInfoUrl": "rest/server/info",
	"restartServiceUrl": "/rest/server/restart",
	"getLogsUrl": "websockets/get_logs",
	"ajaxTimeout": 10000,
	"disableConsole": false
} 